/* devMCA_AIM.c */

/* devMCA_AIM.c - Device Support Routines for Canberra AIM MCA */
/*
 *      Author:  Tim Mooney
 *      Date:    2/25/94
 *
 *      Copyright 1994, the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contract
 *      W-31-109-ENG-38 at Argonne National Laboratory.
 *
 * Modification Log:
 * -----------------
 * .01  02/25/94  tmm  Developed from waveform device support code
 * .02  ??/??/??  MLR  ???
 * .03  09/25/96  tmm  v2.0 minor mods to agree with new version of mca record
 * .03  10/28/96  tmm  v3.0 conversion to EPICS 3.13
 * .04  11/14/97  mlr  V4.0 significant changes to support multiple spectra
 *                          per ADC port, required for use with multiplexors
 *                          and for time-resolved spectroscopy.
 *                          Fixed preset counts feature which had never worked.
 */


#include	<vxWorks.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<stdlib.h>
#include	<wdLib.h>
#include	<memLib.h>
#include	<string.h>

#include	<alarm.h>
#include	<dbDefs.h>
#include	<dbAccess.h>
#include	<recSup.h>
#include	<devSup.h>
#include	<link.h>
/*#include	<rec/dbCommon.h>*/

#include        <nmc_sys_defs.h>
#include	<mcaRecord.h>
#include	"mca.h"

/* Debug support */
#if 0
#define Debug(l,FMT,V) ;
#else
#define Debug(l,FMT,V) {if (l <= devMCA_AIMdebug) \
			  { printf("%s(%d):",__FILE__,__LINE__); \
                            printf(FMT,V);}}
#endif
volatile int devMCA_AIMdebug = 0;

/*
* This structure contains the device specific information for "card", which
* is defined as one ADC port on one AIM
*/
typedef struct {
	long	module;
	long	adc;
	long	base_address;
        long    seq_address;
        long    maxChans;
        long    maxSignals;
        long    maxSequences;
	long	nchans;
	long	plive;
	long	preal;
	long	ptotal;
	long	ptschan;
	long	ptechan;
	long	acqmod;
        long    exists;
} AIM_PortStr;

static long numPorts = 10;
static AIM_PortStr *AIM_Port;

/* Function prototypes */
int AIMSetup(int maxPorts);
int AIMConfig(int card, int address, int port, int maxChans, 
                            int maxSignals, int maxSequences);

/* Create DSET */
static long init_record();
static long send_msg();
static long read_array();
typedef struct {
	long		number;
	DEVSUPFUN	report;
	DEVSUPFUN	init;
	DEVSUPFUN	init_record;
	DEVSUPFUN	get_ioint_info;
	DEVSUPFUN	send_msg;
	DEVSUPFUN	read_array;
} MCA_AIM_DSET;

MCA_AIM_DSET devMCA_AIM = {
	6,
	NULL,
	NULL,
	init_record,
	NULL,
	send_msg,
	read_array
};


/* Routines which are called before iocInit() */

/* Change maximum number of AIM modules from default */
int AIMSetup(int maxPorts)
{
    numPorts = maxPorts;
    return (OK);
}

/* Configure a particular AIM "card" */
int AIMConfig(int card, int address, int port, int maxChans, 
                            int maxSignals, int maxSequences)
{
    unsigned char enet_address[6];
    int s, i;
    AIM_PortStr *p;

    if (AIM_Port == 0) {
        /* This is the first time this routine has been called.  Allocate the
         * AIM_Port array
         */
        AIM_Port = (AIM_PortStr *) malloc(sizeof(AIM_PortStr) * numPorts);
        for (i=0; i<numPorts; i++) {
            AIM_Port[i].exists = 0;
        }
        /* Initialize the AIM software */
        s=nmc_initialize("ei0");
        Debug(5, "(AIMConfig): nmc_initialize=%d\n", s);
    }

    if (card >= numPorts) {
        printf("AIMConfig:  ERROR, card %d exceeds allowed number of cards\n",
                                                    card);
        printf("       Use AIMSetup to increase allowed number.\n");
        return (ERROR);
    }
    p = &AIM_Port[card];
    if (p->exists) {
        printf("AIMConfig:  ERROR, card %d already defined!\n", card);
        return (ERROR);
    }
    p->exists = 1;
    p->maxChans = maxChans;
    p->maxSignals = maxSignals;
    p->maxSequences = maxSequences;
    /* Copy the module ADC port number. Subtract 1 to convert from */
    /* user units to device units  */
    p->adc = port - 1;

    /* Compute the module Ethernet address */
    nmc_build_enet_addr(address, enet_address);

    /* Find the particular module on the net */
    s = nmc_findmod_by_addr(&p->module, enet_address);
    if (s != OK) {
        printf("AIMConfig: ERROR: cannot find module on the network!\n");
        return (ERROR);
    }
    Debug(5, "(AIMConfig): nmc_findmod_by_addr=%d\n", s);
    Debug(5, "(AIMConfig):   module=%d\n", p->module);

    /* Buy the module (make this IOC own it) */
    s = nmc_buymodule(p->module, 0);
    if (s != OK) {
        printf("AIMConfig: ERROR: cannot buy module - someone else owns it!\n");
        return (ERROR);
    }
    Debug(5, "(AIMConfig): nmc_buymodule=%d\n", s);

    /* Allocate memory in the AIM, get base address */
    s = nmc_allocate_memory(p->module,
                    p->maxChans * p->maxSignals * p->maxSequences * 4, 
                    &p->base_address);
    if (s != OK) {
        printf("AIMConfig: ERROR: cannot allocate required memory in AIM\n");
        return (ERROR);
    }
    p->seq_address = p->base_address;
    Debug(5, "(AIMConfig): nmc_allocate_memory=%d\n", s);

    return (OK);
}


/* Support routines */
int send_aim_setup(AIM_PortStr *aim_port)
{
    int s;
    s = nmc_acqu_setup(aim_port->module, aim_port->adc,
		                aim_port->seq_address, 
                                aim_port->maxSignals*aim_port->maxChans,
		                aim_port->plive, aim_port->preal,
		                aim_port->ptotal,
		                aim_port->ptschan, aim_port->ptechan,
		                aim_port->acqmod);
    return(s);
}


int setup_AIM(mcaRecord *pmca)
{
        int card = pmca->inp.value.vmeio.card;
        int signal = pmca->inp.value.vmeio.signal;
	AIM_PortStr *aim_port = AIM_Port + card;
	int s;

	/* Set the module acqusition mode  */
	switch (pmca->mode) {
		case mcaMODE_PHA:   aim_port->acqmod = 1;
			break;
		case mcaMODE_MCS:   aim_port->acqmod = 1;
			break;
		case mcaMODE_List:  aim_port->acqmod = 3;
			break;
	}
	/* Copy the preset real time */
	aim_port->preal = 100 * pmca->prtm;
	/* Copy the preset live time */
	aim_port->plive = 100 * pmca->pltm;
	/* Copy the preset total */
	aim_port->ptotal = pmca->pct;
	/* The preset total start channel is in bytes from start of AIM memory*/
        aim_port->ptschan = aim_port->seq_address + 
                            aim_port->maxChans*signal*4 +
                            pmca->pctl*4;
	/* The preset total end channel is in bytes from start of AIM memory*/
        aim_port->ptechan = aim_port->seq_address + 
                            aim_port->maxChans*signal*4 +
                            pmca->pcth*4;
	/* Copy the number of channels to use */
	aim_port->nchans = pmca->nuse;
	/* Set the address for acquisition and readout */
	aim_port->seq_address = aim_port->base_address + 
                aim_port->maxChans * aim_port->maxSignals * pmca->seq * 4;
	/* Send the parameters to the AIM */
	s = send_aim_setup(aim_port);
	Debug(5, "(setup_AIM): nmc_acqu_setup=%d\n", s);
	return(s);
}


static long init_record(mcaRecord *pmca)
{
        int card = pmca->inp.value.vmeio.card;
        int signal = pmca->inp.value.vmeio.signal;
	AIM_PortStr *aim_port = AIM_Port + card;

	Debug(5, "(init_record): entry\n", 0);
	/* pmca->inp must be VME_IO*/
	if (pmca->inp.type != VME_IO) {
            recGblRecordError(S_db_badField,(void *)pmca,
                "devMCA_AIM (init_record) Illegal INP field");
            return(S_db_badField);
	}

        if (AIM_Port == 0) {
            recGblRecordError(S_db_badField,(void *)pmca,
                "devMCA_AIM (init_record) AIMSetup never called");
            return(S_db_badField);
        }

        if (card >= numPorts) {
            recGblRecordError(S_db_badField,(void *)pmca,
                "devMCA_AIM (init_record) Invalid card number");
            return(S_db_badField);
        }

        if (!aim_port->exists) {
            recGblRecordError(S_db_badField,(void *)pmca,
                "devMCA_AIM (init_record) Card not initialized");
            return(S_db_badField);
        }

        if ((pmca->nuse > aim_port->maxChans) ||
            (pmca->nmax > aim_port->maxChans)) {
            recGblRecordError(S_db_badField,(void *)pmca,
                "devMCA_AIM (init_record) Too many channels");
            return(S_db_badField);
        }

        if (signal >= aim_port->maxSignals) {
            recGblRecordError(S_db_badField,(void *)pmca,
                "devMCA_AIM (init_record) Signal is out of range");
            return(S_db_badField);
        }

        if (pmca->seq >= aim_port->maxSequences) {
            recGblRecordError(S_db_badField,(void *)pmca,
                "devMCA_AIM (init_record) Sequence is out of range");
            return(S_db_badField);
        }

        pmca->nord = 0;
        setup_AIM(pmca);
	return(0);
}

static long send_msg(mcaRecord *pmca, unsigned long msg, void *parg)
{
        int card = pmca->inp.value.vmeio.card;
        int signal = pmca->inp.value.vmeio.signal;
	AIM_PortStr *aim_port = AIM_Port + card;
	long s, elive, ereal, etotals, acq_status, seq, nchans;
	int nuse = pmca->nuse;
        int address;

	aim_port->nchans = nuse;

	switch (msg) {
	case MSG_ACQUIRE:
		/* start acquisition */
		s = nmc_acqu_setstate(aim_port->module, aim_port->adc, 1);
		Debug(5, "(send_msg): start acquisition %d\n", s);
		break;
	case MSG_READ:
		/* start read operation */
		/* This is a no-op. Read-array does everything. */
		break;
	case MSG_SET_CHAS_INT:
		/* set channel advance source to internal (timed) */
		/* This is a NOOP for current MCS hardware - done manually */
		break;
	case MSG_SET_CHAS_EXT:
		/* set channel advance source to external */
		/* This is a NOOP for current MCS hardware - done manually */
		break;
	case MSG_SET_NCHAN:
		/* set number of channels */
                nchans = *(long *)parg;
                if (nchans > aim_port->maxChans) {
                    recGblRecordError(S_db_badField, (void *)pmca,
                    "devMCA_AIM (send_msg) Illegal nuse field");
                    return(S_db_badField);
                }
		aim_port->nchans = nchans;
                s = send_aim_setup(aim_port);
		Debug(5, "(send_msg): acqu setup %d\n", s);
		break;
	case MSG_SET_SEQ:
		/* set sequence number */
		seq = *(long *)parg;
                if (seq > aim_port->maxSequences) {
                    recGblRecordError(S_db_badField, (void *)pmca,
                    "devMCA_AIM (send_msg) Illegal seq field");
                    return(S_db_badField);
                }
	        aim_port->seq_address = aim_port->base_address + 
                    aim_port->maxChans * aim_port->maxSignals * seq * 4;
                s = send_aim_setup(aim_port);
		Debug(5, "(send_msg): acqu setup %d\n", s);
		break;
	case MSG_SET_DWELL:
		/* set dwell time per channel. Convert to centiseconds */
		aim_port->preal = 100 * (*(float *)parg);
                s = send_aim_setup(aim_port);
		Debug(5, "(send_msg): acqu setpresets %d\n", s);
		break;
	case MSG_SET_REAL_TIME:
		/* set preset real time. Convert to centiseconds */
		aim_port->preal = 100 * (*(float *)parg);
                s = send_aim_setup(aim_port);
		Debug(5, "(send_msg): acqu setpresets %d\n", s);
		break;
	case MSG_SET_LIVE_TIME:
		/* set preset live time. Convert to centiseconds */
		aim_port->plive = 100 * (*(float *)parg);
                s = send_aim_setup(aim_port);
		Debug(5, "(send_msg): acqu setpresets %d\n", s);
		break;
	case MSG_SET_COUNTS:
		/* set preset counts */
		aim_port->ptotal = *(long *)parg;
                s = send_aim_setup(aim_port);
		Debug(5, "(send_msg): acqu setpresets %d\n", s);
		break;
	case MSG_SET_LO_CHAN:
		/* set lower side of region integrated for preset counts */
	        /* The preset total start channel is in bytes from start of 
                   AIM memory */
                aim_port->ptschan = aim_port->seq_address + 
                                    aim_port->maxChans*signal*4 +
                                    *(long *)parg*4;
                s = send_aim_setup(aim_port);
		Debug(5, "(send_msg): acqu setpresets %d\n", s);
		break;
	case MSG_SET_HI_CHAN:
		/* set high side of region integrated for preset counts */
	        /* The preset total end channel is in bytes from start of 
                   AIM memory */
                aim_port->ptechan = aim_port->seq_address + 
                                    aim_port->maxChans*signal*4 +
                                    *(long *)parg*4;
                s = send_aim_setup(aim_port);
		Debug(5, "(send_msg): acqu setpresets %d\n", s);
		break;
	case MSG_SET_NSWEEPS:
		/* set number of sweeps (for MCS mode) */
		/* This is a NOOP on current version of MCS */
		break;
	case MSG_SET_MODE_PHA:
		/* set mode to Pulse Height Analysis */
		aim_port->acqmod = 1;
                s = send_aim_setup(aim_port);
		Debug(5, "(send_msg): acqu setup %d\n", s);
		break;
	case MSG_SET_MODE_MCS:
		/* set mode to MultiChannel Scaler */
		aim_port->acqmod = 1;
                s = send_aim_setup(aim_port);
		Debug(5, "(send_msg): acqu setup %d\n", s);
		break;
	case MSG_SET_MODE_LIST:
		/* set mode to LIST (record each incoming event) */
		aim_port->acqmod = 3;
                s = send_aim_setup(aim_port);
		Debug(5, "(send_msg): acqu setup %d\n", s);
		break;
	case MSG_GET_ACQ_STATUS:
		/* Read the current status of the device */
		s = nmc_acqu_statusupdate(aim_port->module, aim_port->adc, 
			0, 0, 0, &elive, &ereal, &etotals, &acq_status);
		Debug(5, "(send_msg): status update %d\n", s);
		pmca->ertm = ereal/100.0;
		pmca->eltm = elive/100.0;
		pmca->act = etotals;
		pmca->acqg = acq_status;
		break;
	case MSG_STOP_ACQUISITION:
		/* stop data acquisition */
		s = nmc_acqu_setstate(aim_port->module, aim_port->adc, 0);
		Debug(5, "(send_msg): stop acquisition %d\n", s);
		break;
	case MSG_ERASE:
		/* erase */
                address = aim_port->seq_address + aim_port->maxChans*signal*4;
		s = nmc_acqu_erase(aim_port->module, address, nuse*4);
		Debug(5, "(send_msg): erase %d\n", s);
                /* Set the elapsed live and real time back to zero. */
                s = nmc_acqu_setelapsed(aim_port->module, aim_port->adc,
                                0, 0);
		Debug(5, "(send_msg): set elapsed %d\n", s);
		break;
	}
	return(0);
}

static long read_array(mcaRecord *pmca)
{
        int card = pmca->inp.value.vmeio.card;
        int signal = pmca->inp.value.vmeio.signal;
	int s, nuse = pmca->nuse;
	long *lptr = (long *)pmca->bptr;
        AIM_PortStr *aim_port = AIM_Port + card;
        int address;

        address = aim_port->seq_address + aim_port->maxChans*signal*4;
	Debug(8, "(read_array): base_address=%x\n", aim_port->base_address);
	Debug(8, "(read_array): seq_address=%x\n", aim_port->seq_address);
	Debug(8, "(read_array): signal=%x\n", signal);
	Debug(8, "(read_array): address=%x\n", address);
	s = nmc_acqu_getmemory_cmp(aim_port->module, aim_port->adc, 
			address, 1, 1, 1, nuse, lptr);
	Debug(5, "(read_array): status=%d\n", s);
	pmca->nord = nuse;
	return(0);
}
