/* mcaRecord.c - Record Support Routines for MCA records */
/*
 *      Original Author: Tim Mooney
 *      Current Author:  Tim Mooney
 *      Date:            2-22-94
 *
 *      Copyright 1991, the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contract
 *      W-31-109-ENG-38 at Argonne National Laboratory.
 *
 * Modification Log:
 * -----------------
 * .01  02-22-94  tmm  Initial development
 * .02  04-03-95  nda  fixed post_monitor on ->bptr
 * .03  07-11-96  tmm  v2.0 added support for regions of interest; allow
 *                     for asynchronous read
 * .04  07-15-96  tmm  v2.2 fixed ROI calcs and posting rules.
 * .05  09-17-96  tmm  v2.3 show ROIs.
 * .06  09-25-96  tmm  v2.4 force read on change of any ROI PV.
 * .06  09-25-96  tmm  v2.5 do misc. ctrl. first instead of last.
 * .07  10-28-96  tmm  v3.0 conversion to EPICS 3.13.
 * .08  11-17-97  mlr  v4.0 Removed ADDR and PORT fields.
 *                          Added SEQ and RMAP fields.
 *                          Removed monitor posting for fields which record
 *                          does not change.
 *                          Put some redundant code in macros.  
 *                          In "process" moved start and stop operations to 
 *                          occur before read status.
 * .09  05-05-98  mlr  v4.1 Changed record to make the timestamp be the time
 *                          when acquisition started, rather than when record
 *                          last processed.
 * .10  09-17-98  mlr  v4.2 Changed logic so ROI changes do not force read,
 *                          since this was too slow when setting many ROIs.
 *                          Instead, compute ROI sums whenever record processes
 *                          since this is a relatively quick operation.
 * .11  09-17-98  tmm  v4.3 Sum ROI's only when needed.  Don't call
 *                          recGblFwdLink() while data acquisition or read is in
 *                          progress, so ca_put_callback() can be used to wait
 *                          for completion of those operations.  Rearranged
 *                          MMAP and NEWV bitmaps.
 * .12  03-17-99  mlr  v4.4 Added PSCL field for prescale factor on MCS
 * .13  03-22-99  mlr  v4.5 Fixed bug in PSCL logic in special
 * .14  04-15-99  mlr  v4.6 Added STIM (start time) field
 * .15  05-07-99  tmm  v4.7 Fixed call to send_msg (only two args)
 */
#define VERSION 4.7

#include	<vxWorks.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<lstLib.h>
#include	<stdlib.h>
#include	<string.h>

#include	<alarm.h>
#include	<dbDefs.h>
#include	<dbAccess.h>
#include	<dbFldTypes.h>
#include	<dbScan.h>
#include	<dbEvent.h>
#include	<devSup.h>
#include	<errMdef.h>
#include	<recSup.h>
#include	<special.h>

#define GEN_SIZE_OFFSET
#include	<mcaRecord.h>
#undef GEN_SIZE_OFFSET
#include	"mca.h"

/* Debug support */
#if 0
#define Debug0(l,FMT) ;
#define Debug1(l,FMT,V) ;
#else
#define Debug0(l,FMT) {if (l <= mcaRecordDebug) \
			  { printf("%s(%d):",__FILE__,__LINE__); \
                            printf(FMT);}}
#define Debug1(l,FMT,V) {if (l <= mcaRecordDebug) \
			  { printf("%s(%d):",__FILE__,__LINE__); \
                            printf(FMT,V);}}
#endif
volatile int mcaRecordDebug = 0;

/* Create RSET - Record Support Entry Table*/
#define report NULL
#define initialize NULL
static long init_record();
static long process();
static long special();
#define get_value NULL
static long cvt_dbaddr();
static long get_array_info();
static long put_array_info();
static long get_units();
static long get_precision();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double();
static long get_control_double();
#define get_alarm_double NULL

struct rset mcaRSET={
	RSETNUMBER,
	report,
	initialize,
	init_record,
	process,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double };

struct mcaDSET { /* mca DSET */
	long            number;
	DEVSUPFUN       dev_report;
	DEVSUPFUN       init;
	DEVSUPFUN       init_record; /*returns: (-1,0)=>(failure,success)*/
	DEVSUPFUN       get_ioint_info;
	DEVSUPFUN       send_msg; /*returns: (-1,0)=>(failure,success)*/
	DEVSUPFUN       read_array; /*returns: (-1,0)=>(failure,success)*/
};

/*sizes of field types (see dbFldTypes.h) */
static int sizeofTypes[] = {0,1,1,2,2,4,4,4,8};

static void monitor();
static long readValue();

#define MAX(a,b) ((a)>(b)?(a):(b))
#define MIN(a,b) ((a)<(b)?(a):(b))

/* Support for Regions Of Interest */
struct roi {
	long lo;
	long hi;
	short nbg;
	unsigned short isPreset;
};
struct roiSum {
	double sum;
	double net;
	double preset;
};
static long sum_ROIs(mcaRecord *pmca, short *preset_reached);
#define NUM_ROI 10

/*******************************************************************************
Support for keeping track of which record fields have been changed, so we can
eliminate redundant db_post_events() without having to think, and without having
to keep lots of "last value of field xxx" fields in the record.  The idea is
to say...

	MARK(M_XXXX);

when you mean...

	db_post_events(pmr, &pmr->xxxx, monitor_mask);

Before leaving, you have to call post_MARKed_fields() to actually post the
field to all listeners.  monitor() does this.

	--- NOTE WELL ---
	The macros below assume that the variable "pmca" exists and points to a
	mca record, like so:
		mcaRecord *pmca;
	No check is made in this code to ensure that this really is true.
*******************************************************************************/

/* These bits are in the mmap field only.  MMAP records fields whose
 * values must be posted.
 */
#define M_VAL		0x00000001
#define M_NACK		0x00000002
#define M_BG		0x00000004
#define M_ACQG		0x00000008
#define	M_READ		0x00000010
#define	M_RDNG		0x00000020
#define	M_ERTM		0x00000040
#define	M_ELTM		0x00000080
#define	M_ACT		0x00000100
#define	M_STRT		0x00000200
#define	M_STOP		0x00000400
#define	M_STIM		0x00000800

/* These bits are in the rmap and newv fields.  RMAP  records ROI fields whose
 * values must be posted.  NEWV records fields whose values are new and for
 * which processing is required.
 */
#define	M_R0		0x00000001
#define	M_R1		0x00000002
#define	M_R2		0x00000004
#define	M_R3		0x00000008
#define	M_R4		0x00000010
#define	M_R5		0x00000020
#define	M_R6		0x00000040
#define	M_R7		0x00000080
#define	M_R8		0x00000100
#define	M_R9		0x00000200
#define M_ROI_ALL	(M_R0|M_R1|M_R2|M_R3|M_R4|M_R5|M_R6|M_R7|M_R8|M_R9)

/* These bits are in the mmap and newv fields */
#define	M_ERAS		0x00000400
#define	M_CHAS		0x00000800
#define	M_NUSE		0x00001000
#define	M_DWEL		0x00002000
#define M_PRTM		0x00004000
#define M_PLTM		0x00008000
#define M_PCT		0x00010000
#define	M_PCTL		0x00020000
#define	M_PCTH		0x00040000
#define	M_PSWP		0x00080000
#define	M_MODE		0x00100000
#define	M_SEQ		0x00200000
#define	M_PSCL		0x00400000


#define MARK(a)		pmca->mmap |= (a);
#define MARKED(a)	(pmca->mmap & (a))
#define UNMARK(a)	pmca->mmap &= ~(a);
#define UNMARK_ALL	pmca->mmap = 0;

#define ROI_MARK(a)	pmca->rmap |= (a);
#define ROI_MARKED(a)	(pmca->rmap & (a))
#define ROI_UNMARK(a)	pmca->rmap &= ~(a);
#define ROI_UNMARK_ALL	pmca->rmap = 0;

#define NEWV_MARK(a)	pmca->newv |= (a);
#define NEWV_MARKED(a)	(pmca->newv & (a))
#define NEWV_UNMARK(a)	pmca->newv &= ~(a);
#define NEWV_UNMARK_ALL	pmca->newv = 0;

/*
state	.acq	.acqg	description
(1)		0		0		quiescent; do nothing
(2)		1		0		user says acquire; send message; if acknowledged,
							go to (3), else set .nack and go to (1).
(3)		1		1		acquiring; do nothing until user or device support
							sends us to (4).
(4)		0		1		acquisition done or halted; initiate read and go to (1).

state	.read	.rdng	description
(1)		0		0		quiescent; do nothing
(2)		1		0		user says "read"; send message (call device support);
                        Device support will send us to (3) if it started an
                        asynchronous read, or leave us in (2) if data are ready.
(3)		1		1		reading; do nothing until user aborts or device support
                        completes, both of which send us to (4).
(4)		0		1		read done or aborted; post and go to (1).
*/

/* The following macro is long, but it saves 6 pages of repetitive code */
#define PROCESS_ROI(DATA_TYPE) \
{\
    DATA_TYPE *pdat = (DATA_TYPE *)pmca->bptr;\
    for (i=0; i<max; i++) ymax = MAX(ymax, pdat[i]);\
}\
    for (i=0; i<NUM_ROI; i++, proi++, psum++) {\
    	DATA_TYPE *p, *pdat = (DATA_TYPE *)pmca->bptr;\
    	DATA_TYPE *pb, *pbg = (DATA_TYPE *)pmca->pbg;\
    	DATA_TYPE *plo, *phi;\
    	sum = net = 0.0;\
    	lo = proi->lo; hi = proi->hi;\
    	if (lo >= 0 && hi >= lo) {\
    	    bg_lo = bg_hi = 0.0;\
    	    if (proi->nbg >= 0) {\
    	    	n = proi->nbg;\
    	    	/* get lo-side background average */\
    	    	plo = &pdat[MAX(lo-n, 0)];\
    	    	phi = &pdat[MIN(lo+n, max)];\
    	    	for (p=plo; p<=phi; p++) bg_lo += *p;\
    	    	bg_lo /= 2*n + 1;\
    	    	/* get hi-side background average */\
    	    	plo = &pdat[MAX(hi-n, 0)];\
    	    	phi = &pdat[MIN(hi+n, max)];\
    	    	for (p=plo; p<=phi; p++) bg_hi += *p;\
    	    	bg_hi /= 2*n + 1;\
    	    }\
    	    /* sum ROI */\
    	    n = hi - lo;\
    	    plo = &pdat[lo];\
    	    for (j=0, p=plo, pb=&pbg[lo]; j<=n; j++, p++, pb++) {\
    	    	sum += *p;\
    	    	if (proi->nbg >= 0)\
    	    	    *pb = bg_lo + (n ? j*(bg_hi-bg_lo)/n : 0); /* linear */\
    	    	net += *p - *pb;\
    	    }\
    	    MARK(M_BG);\
 	}\
 	if ((sum != psum->sum) || (net != psum->net)) ROI_MARK(M_R0<<i);\
	psum->sum = sum;\
	psum->net = net;\
	if (proi->isPreset) *preset_reached |= psum->net >= psum->preset;\
	NEWV_UNMARK(M_R0<<i);\
    }\
\
    proi = (struct roi *)&pmca->r0lo;\
    for (i=0; i<NUM_ROI; i++, proi++) {\
    	DATA_TYPE *pbg = (DATA_TYPE *)pmca->pbg;\
    	if (proi->lo >= 0 && proi->hi >= proi->lo) {\
    	    pbg[proi->lo] = pbg[proi->hi] = ymax;\
   	}\
    }\



static long init_record(mcaRecord *pmca, int pass)
{
	struct mcaDSET *pdset;
	long status;

	/* Allocate memory for spectrum */
	if (pass==0) {
		pmca->vers = VERSION;
		if (pmca->nmax <= 0) pmca->nmax=1;
		if (pmca->ftvl == 0) {
			pmca->bptr = (char *)calloc(pmca->nmax,MAX_STRING_SIZE);
			pmca->pbg = (char *)calloc(pmca->nmax,MAX_STRING_SIZE);
		} else {
			if (pmca->ftvl > DBF_DOUBLE) pmca->ftvl=2;
			pmca->bptr = (char *)calloc(pmca->nmax,sizeofTypes[pmca->ftvl]);
			pmca->pbg = (char *)calloc(pmca->nmax,sizeofTypes[pmca->ftvl]);
		}
		pmca->nord = 0;
		return(0);
	}

	/* simulation links */
	if (pmca->siml.type == CONSTANT) {
		recGblInitConstantLink(&pmca->siml, DBF_USHORT, &pmca->simm);
	}
	if (pmca->siol.type ==	CONSTANT) {
		pmca->nord = 0;
	}

	/* must have dset defined */
	if (!(pdset = (struct mcaDSET *)(pmca->dset))) {
		recGblRecordError(S_dev_noDSET,(void *)pmca,"mca: init_record1");
		return(S_dev_noDSET);
	}
	/* must have read_array function defined */
	if ( (pdset->number < 6) || (pdset->read_array == NULL) ) {
		recGblRecordError(S_dev_missingSup,(void *)pmca,"mca: init_record2");
		printf("%d %p\n",pdset->number, pdset->read_array);
		return(S_dev_missingSup);
	}
	if (pdset->init_record) {
		if ((status=(*pdset->init_record)(pmca))) return(status);
	}
	return(0);
}

static long process(mcaRecord *pmca)
{
	struct mcaDSET *pdset = (struct mcaDSET *)(pmca->dset);
	long status;
	float ertmOld;
	float eltmOld;
	long actOld;
	unsigned short acqgOld;
	short preset_reached = 0;

	/*** Check existence of device support ***/
	if ( (pdset==NULL) || (pdset->read_array==NULL) ) {
		pmca->pact=TRUE;
		recGblRecordError(S_dev_missingSup,(void *)pmca,"read_array");
		return(S_dev_missingSup);
	}

	/* Make sure nobody calls us while we're processing */
	pmca->pact = TRUE;

	/* init Not ACKnowledged flag */
	if (pmca->nack) {pmca->nack = 0; MARK(M_NACK);}


	/*** Handle miscellaneous control fields ***
	* Note that we are allowing these values to be changed even
	* if acquisition is in progress. It is up to device support to
	* prohibit this if necessary.
	*/
	if (pmca->newv && !pmca->rdng) {
		if (NEWV_MARKED(M_CHAS)) {
			MARK(M_CHAS);
			status = (*pdset->send_msg)
				(pmca, (pmca->chas==mcaCHAS_Internal) ?
				MSG_SET_CHAS_INT : MSG_SET_CHAS_EXT, NULL);
			if (status) {pmca->nack = 1; MARK(M_NACK);}
			NEWV_UNMARK(M_CHAS);
		}
		if (NEWV_MARKED(M_NUSE)) {
			MARK(M_NUSE);
			status = (*pdset->send_msg)
				(pmca,  MSG_SET_NCHAN, (void *)(&pmca->nuse));
			if (status) {pmca->nack = 1; MARK(M_NACK);}
			NEWV_UNMARK(M_NUSE);
		}
		if (NEWV_MARKED(M_SEQ)) {
			MARK(M_SEQ);
			status = (*pdset->send_msg)
				(pmca,  MSG_SET_SEQ, (void *)(&pmca->seq));
			if (status) {pmca->nack = 1; MARK(M_NACK);}
			NEWV_UNMARK(M_SEQ);
		}
		if (NEWV_MARKED(M_DWEL)) {
			MARK(M_DWEL);
			status = (*pdset->send_msg)
				(pmca,  MSG_SET_DWELL, (void *)(&pmca->dwel));
			if (status) {pmca->nack = 1; MARK(M_NACK);}
			NEWV_UNMARK(M_DWEL);
		}
		if (NEWV_MARKED(M_PSCL)) {
			MARK(M_PSCL);
			status = (*pdset->send_msg)
				(pmca,  MSG_SET_PSCL, (void *)(&pmca->pscl));
			if (status) {pmca->nack = 1; MARK(M_NACK);}
			NEWV_UNMARK(M_PSCL);
		}
		if (NEWV_MARKED(M_PRTM)) {
			MARK(M_PRTM);
			status = (*pdset->send_msg)
				(pmca,  MSG_SET_REAL_TIME, (void *)(&pmca->prtm));
			if (status) {pmca->nack = 1; MARK(M_NACK);}
			NEWV_UNMARK(M_PRTM);
		}
		if (NEWV_MARKED(M_PLTM)) {
			MARK(M_PLTM);
			status = (*pdset->send_msg)
				(pmca,  MSG_SET_LIVE_TIME, (void *)(&pmca->pltm));
			if (status) {pmca->nack = 1; MARK(M_NACK);}
			NEWV_UNMARK(M_PLTM);
		}
		if (NEWV_MARKED(M_PCT)) {
			MARK(M_PCT);
			status = (*pdset->send_msg)
				(pmca,  MSG_SET_COUNTS, (void *)(&pmca->pct));
			if (status) {pmca->nack = 1; MARK(M_NACK);}
			NEWV_UNMARK(M_PCT);
		}
		if (NEWV_MARKED(M_PCTL)) {
			MARK(M_PCTL);
			status = (*pdset->send_msg)
				(pmca,  MSG_SET_LO_CHAN, (void *)(&pmca->pctl));
			if (status) {pmca->nack = 1; MARK(M_NACK);}
			NEWV_UNMARK(M_PCTL);
		}
		if (NEWV_MARKED(M_PCTH)) {
			MARK(M_PCTH);
			status = (*pdset->send_msg)
				(pmca,  MSG_SET_HI_CHAN, (void *)(&pmca->pcth));
			if (status) {pmca->nack = 1; MARK(M_NACK);}
			NEWV_UNMARK(M_PCTH);
		}
		if (NEWV_MARKED(M_PSWP)) {
			MARK(M_PSWP);
			status = (*pdset->send_msg)
				(pmca,  MSG_SET_NSWEEPS, (void *)(&pmca->pswp));
			if (status) {pmca->nack = 1; MARK(M_NACK);}
			NEWV_UNMARK(M_PSWP);
		}
		if (NEWV_MARKED(M_ERAS)) {
			MARK(M_ERAS);
			status = (*pdset->send_msg)
				(pmca,  MSG_ERASE, NULL);
			if (status) {pmca->nack = 1; MARK(M_NACK);}
			NEWV_UNMARK(M_ERAS);
			pmca->eras = 0;
			/* Use TimeStamp to record beginning of acquisition */
			recGblGetTimeStamp(pmca);
                        tsStampToText(&pmca->time, TS_TEXT_MONDDYYYY,
                                                   pmca->stim);
                        MARK(M_STIM);
		}
		if (NEWV_MARKED(M_MODE)) {
			MARK(M_MODE);
			switch (pmca->mode) {
			case mcaMODE_PHA:
			default:
				status = (*pdset->send_msg) (pmca, MSG_SET_MODE_PHA, NULL);
				break;
			case mcaMODE_MCS:
				status = (*pdset->send_msg) (pmca, MSG_SET_MODE_MCS, NULL);
				break;
			case mcaMODE_List:
				status = (*pdset->send_msg) (pmca, MSG_SET_MODE_LIST, NULL);
				break;
			}
			if (status) {pmca->nack = 1; MARK(M_NACK);}
			NEWV_UNMARK(M_MODE);
		}
	}

	/* Turn acquisition on or off.  Do this before reading device status */
	if (pmca->strt) {
		Debug0(5,"process: start acquisition.\n");
		status = (*pdset->send_msg) (pmca, MSG_ACQUIRE, NULL);
		pmca->strt=0; MARK(M_STRT);
		/* Use TimeStamp to record beginning of acquisition */
		/* Don't reset the clock if we are already acquiring */
		if (!pmca->acqg) {
                    recGblGetTimeStamp(pmca);
                    tsStampToText(&pmca->time, TS_TEXT_MONDDYYYY,
                                                   pmca->stim);
                    MARK(M_STIM);
                }
	}
	if (pmca->stop) {
		Debug0(5,"process: stop acquisition.\n");
		status = (*pdset->send_msg) (pmca, MSG_STOP_ACQUISITION, NULL);
		pmca->stop=0; MARK(M_STOP);
	}


	/*** Read the device status. ***
	* This includes the elapsed live time, elapsed real time, actual counts
	* in preset region and the current acquisition status. 
	*/

	/* Save fields updated by MSG_GET_ACQ_STATUS */
	acqgOld = pmca->acqg;
	ertmOld = pmca->ertm;
	eltmOld = pmca->eltm;
	actOld = pmca->act;

	Debug0(5,"process: reading elapsed time, etc.\n");
	status = (*pdset->send_msg)(pmca, MSG_GET_ACQ_STATUS, NULL);
	if (status) {
		pmca->nack = 1; MARK(M_NACK);
		pmca->acqg = 0;
	}

	if (pmca->ertm != ertmOld) MARK(M_ERTM);
	if (pmca->eltm != eltmOld) MARK(M_ELTM);
	if (pmca->act != actOld)   MARK(M_ACT);

	if (pmca->acqg != acqgOld) {
		MARK(M_ACQG);
		/* Force a read when acquire turns off */
		if (!pmca->acqg) {
			pmca->read = 1; MARK(M_READ);
		}
	}


	/*** Handle read cycle ***/
	if (pmca->read != pmca->rdng) {
		/* either start a read or handle read completed/aborted */
		if (pmca->read) {
			/* start a read */
			Debug0(5,"process: initiate read.\n");
			status = (*pdset->send_msg)(pmca, MSG_READ, NULL);
			if (status) {
				/* no ack from device support; go quiescent */
				pmca->nack = 1; MARK(M_NACK);
				pmca->read = 0; MARK(M_READ);
			} else {
				/*
				 * If read requires a callback, device support will tell us so
				 * by setting pmca->rdng = 1, and we'll do nothing until we get
				 * processed again.  Otherwise, we can get the data immediately.
				 */
				if (pmca->rdng) {
					MARK(M_RDNG);
				} else {
					/* get data */
					Debug0(5,"process: get data (no callback).\n");
					status = readValue(pmca); /* read the new value */
					if (status) {
						pmca->nack = 1; MARK(M_NACK);
					} else {
						MARK(M_VAL);
					}
					pmca->read = 0; MARK(M_READ);
				}
			}
		} else {
			/* handle read completed/aborted */
			Debug0(5,"process: get data (callback or aborted read).\n");
			status = readValue(pmca); /* read the new value */
			if (status) {
				pmca->nack = 1; MARK(M_NACK);
			} else {
				MARK(M_VAL);
			}
			pmca->rdng = 0; MARK(M_RDNG);
		}

	}

	/* If any ROI is marked, sumROIs */
	if (NEWV_MARKED(M_ROI_ALL)) {
        (void)sum_ROIs(pmca, &preset_reached);
	}

	if (preset_reached) {
		Debug0(5,"process: stop acquisition.\n");
		status = (*pdset->send_msg) (pmca, MSG_STOP_ACQUISITION, NULL);
	}

	pmca->udf = FALSE;

	monitor(pmca);

	/*
	 * Process forward-linked record.  Tell EPICS dbPutNotify mechanism
	 * that processing is finished.
	 */
	if (!pmca->acqg && !pmca->rdng) recGblFwdLink(pmca);

	pmca->pact=FALSE;
	return(0);
}

static long cvt_dbaddr(struct dbAddr *paddr)
{
	mcaRecord *pmca=(mcaRecord *)paddr->precord;
	int fieldIndex = dbGetFieldIndex(paddr);

	if (fieldIndex == mcaRecordVAL) {
		paddr->pfield = (void *)(pmca->bptr);
	} else if (fieldIndex == mcaRecordBG) {
		paddr->pfield = (void *)(pmca->pbg);
	}
	paddr->no_elements = pmca->nmax;
	paddr->field_type = pmca->ftvl;
	if (pmca->ftvl==0)  paddr->field_size = MAX_STRING_SIZE;
	else paddr->field_size = sizeofTypes[pmca->ftvl];
	paddr->dbr_field_type = pmca->ftvl;
	return(0);
}

static long get_array_info(struct dbAddr *paddr, long *no_elements, long *offset)
{
	mcaRecord *pmca=(mcaRecord *)paddr->precord;

	*no_elements =  pmca->nord;
	*offset = 0;
	return(0);
}

static long put_array_info(struct dbAddr *paddr, long nNew)
{
	mcaRecord *pmca=(mcaRecord *)paddr->precord;

	pmca->nord = nNew;
	if (pmca->nord > pmca->nmax) pmca->nord = pmca->nmax;
	return(0);
}

static long get_units(struct dbAddr *paddr, char *units)
{
	mcaRecord *pmca=(mcaRecord *)paddr->precord;

	strncpy(units,pmca->egu,sizeof(pmca->egu));
	return(0);
}

static long get_precision(struct dbAddr *paddr, long *precision)
{
	mcaRecord *pmca=(mcaRecord *)paddr->precord;

	*precision = pmca->prec;
	if (paddr->pfield==(void *)pmca->bptr) return(0);
	recGblGetPrec(paddr,precision);
	return(0);
}

static long get_graphic_double(struct dbAddr *paddr, struct dbr_grDouble *pgd)
{
	mcaRecord *pmca=(mcaRecord *)paddr->precord;
	int fieldIndex = dbGetFieldIndex(paddr);

	if (fieldIndex == mcaRecordBPTR) {
		pgd->upper_disp_limit = pmca->hopr;
		pgd->lower_disp_limit = pmca->lopr;
	} else recGblGetGraphicDouble(paddr,pgd);
	return(0);
}
static long get_control_double(paddr,pcd)
struct dbAddr *paddr;
struct dbr_ctrlDouble *pcd;
{
	mcaRecord *pmca=(mcaRecord *)paddr->precord;
	int fieldIndex = dbGetFieldIndex(paddr);

	if (fieldIndex == mcaRecordBPTR) {
		pcd->upper_ctrl_limit = pmca->hopr;
		pcd->lower_ctrl_limit = pmca->lopr;
	} else recGblGetControlDouble(paddr,pcd);
	return(0);
}

static void monitor(mcaRecord *pmca)
{
	unsigned short monitor_mask;

	monitor_mask = recGblResetAlarms(pmca);
	monitor_mask |= (DBE_VALUE|DBE_LOG);
	if (MARKED(M_VAL)) db_post_events(pmca,pmca->bptr,monitor_mask);
	if (MARKED(M_BG))   db_post_events(pmca,pmca->pbg,monitor_mask);
	if (MARKED(M_NACK)) db_post_events(pmca,&pmca->nack,monitor_mask);
	if (MARKED(M_ACQG)) db_post_events(pmca,&pmca->acqg,monitor_mask);
	if (MARKED(M_READ)) db_post_events(pmca,&pmca->read,monitor_mask);
	if (MARKED(M_RDNG)) db_post_events(pmca,&pmca->rdng,monitor_mask);
	if (MARKED(M_ERTM)) db_post_events(pmca,&pmca->ertm,monitor_mask);
	if (MARKED(M_ELTM)) db_post_events(pmca,&pmca->eltm,monitor_mask);
	if (MARKED(M_ACT )) db_post_events(pmca,&pmca->act ,monitor_mask);
	if (MARKED(M_STRT)) db_post_events(pmca,&pmca->strt,monitor_mask);
	if (MARKED(M_STOP)) db_post_events(pmca,&pmca->stop,monitor_mask);
	if (MARKED(M_STIM)) db_post_events(pmca,&pmca->stim,monitor_mask);

	if (ROI_MARKED(M_R0)) {
		db_post_events(pmca,&pmca->r0 ,monitor_mask);
		db_post_events(pmca,&pmca->r0n ,monitor_mask);
	}
	if (ROI_MARKED(M_R1)) {
		db_post_events(pmca,&pmca->r1 ,monitor_mask);
		db_post_events(pmca,&pmca->r1n ,monitor_mask);
	}
	if (ROI_MARKED(M_R2)) {
		db_post_events(pmca,&pmca->r2 ,monitor_mask);
		db_post_events(pmca,&pmca->r2n ,monitor_mask);
	}
	if (ROI_MARKED(M_R3)) {
		db_post_events(pmca,&pmca->r3 ,monitor_mask);
		db_post_events(pmca,&pmca->r3n ,monitor_mask);
	}
	if (ROI_MARKED(M_R4)) {
		db_post_events(pmca,&pmca->r4 ,monitor_mask);
		db_post_events(pmca,&pmca->r4n ,monitor_mask);
	}
	if (ROI_MARKED(M_R5)) {
		db_post_events(pmca,&pmca->r5 ,monitor_mask);
		db_post_events(pmca,&pmca->r5n ,monitor_mask);
	}
	if (ROI_MARKED(M_R6)) {
		db_post_events(pmca,&pmca->r6 ,monitor_mask);
		db_post_events(pmca,&pmca->r6n ,monitor_mask);
	}
	if (ROI_MARKED(M_R7)) {
		db_post_events(pmca,&pmca->r7 ,monitor_mask);
		db_post_events(pmca,&pmca->r7n ,monitor_mask);
	}
	if (ROI_MARKED(M_R8)) {
		db_post_events(pmca,&pmca->r8 ,monitor_mask);
		db_post_events(pmca,&pmca->r8n ,monitor_mask);
	}
	if (ROI_MARKED(M_R9)) {
		db_post_events(pmca,&pmca->r9 ,monitor_mask);
		db_post_events(pmca,&pmca->r9n ,monitor_mask);
	}
	UNMARK_ALL;
	ROI_UNMARK_ALL;
	return;

}

static long special(struct dbAddr *paddr, int after)
{
	mcaRecord *pmca=(mcaRecord *)paddr->precord;
	int i, fieldIndex = dbGetFieldIndex(paddr);

	if (!after) return(0);

	/* new dir */
	switch (fieldIndex) {
	case mcaRecordERAS:	NEWV_MARK(M_ERAS); break;
	case mcaRecordCHAS:	NEWV_MARK(M_CHAS); break;
	case mcaRecordNUSE:	NEWV_MARK(M_NUSE); break;
	case mcaRecordSEQ:	NEWV_MARK(M_SEQ); break;
	case mcaRecordDWEL:	NEWV_MARK(M_DWEL); break;
	case mcaRecordPSCL:	NEWV_MARK(M_PSCL); break;
	case mcaRecordPRTM:	NEWV_MARK(M_PRTM); break;
	case mcaRecordPLTM:	NEWV_MARK(M_PLTM); break;
	case mcaRecordPCT:	NEWV_MARK(M_PCT); break;
	case mcaRecordPCTL:	NEWV_MARK(M_PCTL); break;
	case mcaRecordPCTH:	NEWV_MARK(M_PCTH); break;
	case mcaRecordPSWP:	NEWV_MARK(M_PSWP); break;
	case mcaRecordMODE:	NEWV_MARK(M_MODE); break;
	default:
		if ((fieldIndex >= mcaRecordR0LO) && (fieldIndex <= mcaRecordR9IP)) {
			/* Which ROI is affected? */
			i = (fieldIndex - mcaRecordR0LO)/4;
			/* Mark the ROI for recalculation. */
			NEWV_MARK(M_R0 << i);
		}
		break;
	}

	/* Do we need to read the AIM after command gets executed?
	 * Ideally we would like to force a read after changing number of
	 * channels, erasing or changing sequence.  Erase is typically slow 
	 * so forcing a read does not really hurt performance much.  Changing
	 * number of channels is typically done infrequently so forcing a read
	 * is OK in that case too.  However, changing sequence is typically
	 * done for time-resolved spectroscopy, so performance really matters.
	 * Thus, we don't force a read on change of seq field.
	 */
	if (NEWV_MARKED(M_ERAS) || NEWV_MARKED(M_NUSE)) {
		pmca->read = 1;
		MARK(M_READ);
	}
	return(0);
}

static long readValue(mcaRecord *pmca)
{
	long status;
	struct mcaDSET *pdset = (struct mcaDSET *) (pmca->dset);
	long nRequest = 1;

	NEWV_MARK(M_ROI_ALL);  /* Mark all ROI's for ercalculation. */

	status = dbGetLink(&(pmca->siml), DBR_ENUM, &(pmca->simm), NULL, NULL);
	if (status) return(status);

	if (pmca->simm == NO) {
		status = (*pdset->read_array)(pmca);
		return(status);
	}
	if (pmca->simm == YES) {
		nRequest = pmca->nmax;
		status = dbGetLink(&(pmca->siol), pmca->ftvl, pmca->bptr,NULL, &nRequest);
		/* nord set only for db links: needed for old db_access */
		if (pmca->siol.type == DB_LINK) pmca->nord = nRequest;
		if (status == 0) {
			pmca->udf = FALSE;
		}
	} else {
		status=-1;
		recGblSetSevr(pmca,SOFT_ALARM,INVALID_ALARM);
		return(status);
	}
	recGblSetSevr(pmca,SIMM_ALARM,pmca->sims);

	return(status);
}


static long sum_ROIs(mcaRecord *pmca, short *preset_reached)
{
	int i, j, n, max, lo, hi;
	double sum, bg_lo, bg_hi, net, ymax=0.0;
	struct roi *proi = (struct roi *)&pmca->r0lo;
	struct roiSum *psum = (struct roiSum *)&pmca->r0;

	Debug0(5,"sum_ROIs: entry\n");
	(void)memset(pmca->pbg, 0, pmca->nmax*sizeofTypes[pmca->ftvl]);
	*preset_reached = 0;
	max = pmca->nord-1;

	switch (pmca->ftvl) {
	case DBF_CHAR:
	case DBF_UCHAR:
	default:
		break;

	case DBF_SHORT:
		{PROCESS_ROI(short);
		 break;
		}
	case DBF_USHORT:
		{PROCESS_ROI(unsigned short);
		 break;
		}
	case DBF_LONG:
		{PROCESS_ROI(long);
		 break;
		}
	case DBF_ULONG:
		{PROCESS_ROI(unsigned long);
		 break;
		}

	case DBF_FLOAT:
		{PROCESS_ROI(float);
		 break;
		}
	case DBF_DOUBLE:
		{PROCESS_ROI(double);
		 break;
		}
	}
	return(0);
}
